import logging
import traceback

from getpass import getpass

from moodle_dl.utils import cutie
from moodle_dl.utils.logger import Log
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.url_target import URLTarget
from moodle_dl.notification_services.xmpp.xmpp_shooter import XmppShooter
from moodle_dl.notification_services.notification_service import NotificationService
from moodle_dl.notification_services.xmpp.xmpp_formater import XmppFormater as XF


class XmppService(NotificationService):
    def interactively_configure(self) -> None:
        """
        Guides the user through the configuration of the xmpp notification.
        """

        do_xmpp = cutie.prompt_yes_or_no('Do you want to activate Notifications via XMPP?')

        if not do_xmpp:
            self.config_helper.remove_property('xmpp')
        else:
            print('[The following Inputs are not validated!]')
            config_valid = False
            while not config_valid:
                sender = input('JID of the Sender:   ')
                password = getpass('Password for the Sender [no output]:   ')
                target = input('JID of the Target:   ')
                print('Testing XMPP-Config...')

                try:
                    xmpp_shooter = XmppShooter(sender, password, target)
                    xmpp_shooter.send('This is a Testmessage from Moodle Downloader!')
                except BaseException as e:
                    print('Error while sending the test message: %s' % (str(e)))
                    continue

                else:
                    input(
                        'Please check if you received the Testmessage.'
                        + ' If yes, confirm with Return.\nIf not, exit'
                        + ' this program ([CTRL]+[C]) and try again later.'
                    )
                    config_valid = True

                raw_send_error_msg = ''
                while raw_send_error_msg not in ['y', 'n']:
                    raw_send_error_msg = input('Do you want to also get error reports sent in xmpp? [y/n]   ')

                do_send_error_msg = raw_send_error_msg == 'y'

                xmpp_cfg = {
                    'sender': sender,
                    'password': password,
                    'target': target,
                    'send_error_msg': do_send_error_msg,
                }

                self.config_helper.set_property('xmpp', xmpp_cfg)

    def _is_configured(self) -> bool:
        # Checks if the sending of XMPP messages has been configured.
        try:
            self.config_helper.get_property('xmpp')
            return True
        except ValueError:
            logging.debug('XMPP-Notifications not configured, skipping.')
            return False

    def _send_messages(self, messages: [str]):
        """
        Sends an message
        """
        if not self._is_configured() or messages is None or len(messages) == 0:
            return

        xmpp_cfg = self.config_helper.get_property('xmpp')

        logging.debug('Sending Notification via XMPP...')
        Log.debug('Sending Notification via XMPP... (Please wait)')

        try:
            xmpp = XmppShooter(xmpp_cfg['sender'], xmpp_cfg['password'], xmpp_cfg['target'])
            xmpp.send_messages(messages)
        except BaseException as e:
            error_formatted = traceback.format_exc()
            logging.error('While sending notification:\n%s' % (error_formatted), extra={'exception': e})
            raise e  # to be properly notified via Sentry

    def notify_about_changes_in_moodle(self, changes: [Course]) -> None:
        """
        Sends out a notification about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        if not self._is_configured():
            return

        messages = XF.create_full_moodle_diff_messages(changes)

        self._send_messages(messages)

    def notify_about_error(self, error_description: str):
        """
        Sends out an error message if configured to do so.
        @param error_description: The error object.
        """
        if not self._is_configured():
            return

        xmpp_cfg = self.config_helper.get_property('xmpp')

        if not xmpp_cfg.get('send_error_msg', True):
            return
        messages = XF.create_full_error_messages(error_description)

        self._send_messages(messages)

    def notify_about_failed_downloads(self, failed_downloads: [URLTarget]):
        """
        Sends out an message about failed download if configured to send out error messages.
        @param failed_downloads: A list of failed URLTargets.
        """
        if not self._is_configured():
            return

        xmpp_cfg = self.config_helper.get_property('xmpp')

        if not xmpp_cfg.get('send_error_msg', True):
            return
        messages = XF.create_full_failed_downloads_messages(failed_downloads)

        self._send_messages(messages)



from moodle_dl.moodle_connector.request_helper import RequestHelper
from moodle_dl.state_recorder.course import Course


class FirstContactHandler:
    """
    Fetches and parses the various endpoints in Moodle.
    """

    def __init__(self, request_helper: RequestHelper):
        self.request_helper = request_helper
        # oldest supported Moodle version
        self.version = 2011120500

    def fetch_userid_and_version(self) -> str:
        """
        Ask the Moodle system for the user id.
        @return: the userid
        """
        result = self.request_helper.post_REST('core_webservice_get_site_info')

        if 'userid' not in result:
            raise RuntimeError('Error could not receive your user ID!')
        userid = result.get('userid', '')

        version = result.get('version', '2011120500')

        try:
            version = int(version.split('.')[0])
        except Exception as e:
            raise RuntimeError('Error could not parse version string: "%s" Error: %s' % (version, e))

        self.version = version
        return userid, version

    def fetch_courses(self, userid: str) -> [Course]:
        """
        Queries the Moodle system for all courses the user
        is enrolled in.
        @param userid: the user id
        @return: A list of courses
        """
        data = {'userid': userid}

        result = self.request_helper.post_REST('core_enrol_get_users_courses', data)

        results = []
        for course in result:
            results.append(Course(course.get('id', 0), course.get('fullname', '')))
        return results

    def fetch_courses_info(self, course_ids: [int]) -> [Course]:
        """
        Queries the Moodle system for info about courses in a list.
        @param course_ids: A list of courses ids
        @return: A list of courses
        """
        # API is only available since version 3.2
        if len(course_ids) == 0 or self.version < 2016120500:
            return []

        data = {
            "field": "ids",
            "value": ",".join(list(map(str, course_ids))),
        }

        result = self.request_helper.post_REST('core_course_get_courses_by_field', data)
        result = result.get('courses', [])

        results = []
        for course in result:
            results.append(Course(course.get('id', 0), course.get('fullname', '')))
        return results



from youtube_dl.YoutubeDL import YoutubeDL

from moodle_dl.ydl_extractors.zoomus import ZoomUSIE
from moodle_dl.ydl_extractors.opencast import OpencastIE, OpencastPlaylistIE
from moodle_dl.ydl_extractors.helixmedia import Helixmedia
from moodle_dl.ydl_extractors.owncloud import Owncloud
from moodle_dl.ydl_extractors.opencast_lti import OpencastLTI
from moodle_dl.ydl_extractors.googledrive import GoogleDriveIE


def add_additional_extractors(ydl: YoutubeDL):
    additional_extractors = [
        OpencastIE(ydl),
        OpencastPlaylistIE(ydl),
        ZoomUSIE(ydl),
        Helixmedia(ydl),
        Owncloud(ydl),
        OpencastLTI(ydl),
        GoogleDriveIE(ydl),
    ]

    for extractor in additional_extractors:
        ydl._ies.insert(0, extractor)
        ydl._ies_instances[extractor.ie_key()] = extractor



from moodle_dl.notification_services.telegram.telegram_formater import TelegramFormater


class XmppFormater(TelegramFormater):
    @staticmethod
    def append_with_limit(new_line: str, one_msg_content: str, msg_list: [str]):
        """Appends a new line to a message string,
        if the string is to long it ist appended to the message list.
        Returns the new message string.

        Args:
            new_line (str): A new line to append to a message string
            one_msg_content (str): The current message string
            msg_list ([str]): The list of finished messages
        Returns:
            str: The new message
        """
        if len(one_msg_content) + len(new_line) >= 4096:
            msg_list.append(one_msg_content)
            return new_line
        else:
            return one_msg_content + new_line



# coding: utf-8
from __future__ import unicode_literals


import re
import json

from youtube_dl.compat import (
    compat_urllib_parse,
    compat_urllib_parse_urlparse,
)

from youtube_dl.extractor.common import InfoExtractor
from youtube_dl.utils import (
    ExtractorError,
    urlencode_postdata,
    js_to_json,
    determine_ext,
    mimetype2ext,
    extract_attributes,
    HEADRequest,
)


class OpencastLTI(InfoExtractor):
    IE_NAME = 'opencastLTI'
    _VALID_URL = r'(?P<scheme>https?://)(?P<host>[^/]+)(?P<path>.*)?/mod/lti/view.php\?.*?id=(?P<id>\d+)'
    _LAUNCH_FORM = 'ltiLaunchForm'

    # _TEST = {'url': 'http://moodle.ruhr-uni-bochum.de/mod/lti/view.php?id=1406269'}

    def _real_extract(self, url):
        mobj = re.match(self._VALID_URL, url)
        scheme = mobj.group('scheme')
        host = mobj.group('host')
        path = mobj.group('path')
        video_id = mobj.group('id')

        launch_url = scheme + host + path + '/mod/lti/launch.php?id=' + video_id

        # webpage = self._download_webpage(url, video_id)
        launch_webpage = self._download_webpage(launch_url, video_id, 'Downloading opencast lti launch webpage')
        launch_inputs = self._form_hidden_inputs(self._LAUNCH_FORM, launch_webpage)
        launch_form_str = self._search_regex(
            r'(?P<form><form[^>]+?id=(["\'])%s\2[^>]*>)' % self._LAUNCH_FORM, launch_webpage, 'login form', group='form'
        )

        action_url = extract_attributes(launch_form_str).get('action')

        submit_page, start_urlh = self._download_webpage_handle(
            action_url, video_id, 'Launch opencast app', data=urlencode_postdata(launch_inputs)
        )

        if start_urlh.status != 200:
            raise ExtractorError('Unable to launch opencast app', expected=True)

        return {
            '_type': 'url',
            'url': start_urlh.geturl(),
        }



from moodle_dl.state_recorder.file import File
from moodle_dl.download_service.path_tools import PathTools


class Course:
    def __init__(self, _id: int, fullname: str, files: [File] = None):
        self.id = _id
        self.fullname = fullname
        if files is not None:
            self.files = files
        else:
            self.files = []

        self.overwrite_name_with = None
        self.create_directory_structure = True

    def __str__(self):
        message = 'Course ('

        message += 'id: %s' % (self.id)
        message += ', fullname: "%s"' % (PathTools.to_valid_name(self.fullname))
        message += ', overwrite_name_with: "%s"' % (PathTools.to_valid_name(self.overwrite_name_with))
        message += ', create_directory_structure: %s' % (self.create_directory_structure)
        message += ', files: %s' % (len(self.files))

        # for i, file in enumerate(self.files):
        #     message += ', file[%i]: %s' % (i, file)

        message += ')'
        return message



moodle_html_header = '''
<!DOCTYPE html>

<html dir="ltr" lang="de" xml:lang="de">

<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <link rel="stylesheet" href="http://maxcdn.bootstrapcdn.com/font-awesome/latest/css/font-awesome.min.css">
    <script id="firstthemesheet"
        type="text/css">/** Required in order to fix style inclusion problems in IE with YUI **/</script>
    <link rel="stylesheet" type="text/css"
        href="https://moodle.ruhr-uni-bochum.de/theme/styles.php/boost/1633802793_1/all" />
    <script>
        //<![CDATA[
        var M = {}; M.yui = {};
        M.pageloadstarttime = new Date();
        M.cfg = { "wwwroot": "https:\\/\\/moodle.ruhr-uni-bochum.de", "sesskey": "roflAffe", "sessiontimeout": "28800", "themerev": "1633802793", "slasharguments": 1, "theme": "boost", "iconsystemmodule": "core\\/icon_system_fontawesome", "jsrev": "1633802793", "admin": "admin", "svgicons": true, "usertimezone": "Europa\\/London", "contextid": 104, "langrev": 1633946764, "templaterev": "1633802793" }; var yui1ConfigFn = function (me) { if (/-skin|reset|fonts|grids|base/.test(me.name)) { me.type = 'css'; me.path = me.path.replace(/\\.js/, '.css'); me.path = me.path.replace(/\\/yui2-skin/, '/assets/skins/sam/yui2-skin') } };
        var yui2ConfigFn = function (me) {
            var parts = me.name.replace(/^moodle-/, '').split('-'), component = parts.shift(), module = parts[0], min = '-min'; if (/-(skin|core)$/.test(me.name)) { parts.pop(); me.type = 'css'; min = '' }
            if (module) { var filename = parts.join('-'); me.path = component + '/' + module + '/' + filename + min + '.' + me.type } else { me.path = component + '/' + component + '.' + me.type }
        };
        YUI_config = { "debug": false, "base": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/yuilib\\/3.17.2\\/", "comboBase": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/yui_combo.php?", "combine": true, "filter": null, "insertBefore": "firstthemesheet", "groups": { "yui2": { "base": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/yuilib\\/2in3\\/2.9.0\\/build\\/", "comboBase": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/yui_combo.php?", "combine": true, "ext": false, "root": "2in3\\/2.9.0\\/build\\/", "patterns": { "yui2-": { "group": "yui2", "configFn": yui1ConfigFn } } }, "moodle": { "name": "moodle", "base": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/yui_combo.php?m\\/1633802793\\/", "combine": true, "comboBase": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/yui_combo.php?", "ext": false, "root": "m\\/1633802793\\/", "patterns": { "moodle-": { "group": "moodle", "configFn": yui2ConfigFn } }, "filter": null, "modules": { "moodle-core-event": { "requires": ["event-custom"] }, "moodle-core-chooserdialogue": { "requires": ["base", "panel", "moodle-core-notification"] }, "moodle-core-languninstallconfirm": { "requires": ["base", "node", "moodle-core-notification-confirm", "moodle-core-notification-alert"] }, "moodle-core-blocks": { "requires": ["base", "node", "io", "dom", "dd", "dd-scroll", "moodle-core-dragdrop", "moodle-core-notification"] }, "moodle-core-tooltip": { "requires": ["base", "node", "io-base", "moodle-core-notification-dialogue", "json-parse", "widget-position", "widget-position-align", "event-outside", "cache-base"] }, "moodle-core-formchangechecker": { "requires": ["base", "event-focus", "moodle-core-event"] }, "moodle-core-dragdrop": { "requires": ["base", "node", "io", "dom", "dd", "event-key", "event-focus", "moodle-core-notification"] }, "moodle-core-popuphelp": { "requires": ["moodle-core-tooltip"] }, "moodle-core-lockscroll": { "requires": ["plugin", "base-build"] }, "moodle-core-handlebars": { "condition": { "trigger": "handlebars", "when": "after" } }, "moodle-core-actionmenu": { "requires": ["base", "event", "node-event-simulate"] }, "moodle-core-notification": { "requires": ["moodle-core-notification-dialogue", "moodle-core-notification-alert", "moodle-core-notification-confirm", "moodle-core-notification-exception", "moodle-core-notification-ajaxexception"] }, "moodle-core-notification-dialogue": { "requires": ["base", "node", "panel", "escape", "event-key", "dd-plugin", "moodle-core-widget-focusafterclose", "moodle-core-lockscroll"] }, "moodle-core-notification-alert": { "requires": ["moodle-core-notification-dialogue"] }, "moodle-core-notification-confirm": { "requires": ["moodle-core-notification-dialogue"] }, "moodle-core-notification-exception": { "requires": ["moodle-core-notification-dialogue"] }, "moodle-core-notification-ajaxexception": { "requires": ["moodle-core-notification-dialogue"] }, "moodle-core-maintenancemodetimer": { "requires": ["base", "node"] }, "moodle-core_availability-form": { "requires": ["base", "node", "event", "event-delegate", "panel", "moodle-core-notification-dialogue", "json"] }, "moodle-backup-backupselectall": { "requires": ["node", "event", "node-event-simulate", "anim"] }, "moodle-backup-confirmcancel": { "requires": ["node", "node-event-simulate", "moodle-core-notification-confirm"] }, "moodle-course-util": { "requires": ["node"], "use": ["moodle-course-util-base"], "submodules": { "moodle-course-util-base": {}, "moodle-course-util-section": { "requires": ["node", "moodle-course-util-base"] }, "moodle-course-util-cm": { "requires": ["node", "moodle-course-util-base"] } } }, "moodle-course-management": { "requires": ["base", "node", "io-base", "moodle-core-notification-exception", "json-parse", "dd-constrain", "dd-proxy", "dd-drop", "dd-delegate", "node-event-delegate"] }, "moodle-course-categoryexpander": { "requires": ["node", "event-key"] }, "moodle-course-formatchooser": { "requires": ["base", "node", "node-event-simulate"] }, "moodle-course-dragdrop": { "requires": ["base", "node", "io", "dom", "dd", "dd-scroll", "moodle-core-dragdrop", "moodle-core-notification", "moodle-course-coursebase", "moodle-course-util"] }, "moodle-form-passwordunmask": { "requires": [] }, "moodle-form-dateselector": { "requires": ["base", "node", "overlay", "calendar"] }, "moodle-form-shortforms": { "requires": ["node", "base", "selector-css3", "moodle-core-event"] }, "moodle-question-searchform": { "requires": ["base", "node"] }, "moodle-question-chooser": { "requires": ["moodle-core-chooserdialogue"] }, "moodle-question-preview": { "requires": ["base", "dom", "event-delegate", "event-key", "core_question_engine"] }, "moodle-availability_completion-form": { "requires": ["base", "node", "event", "moodle-core_availability-form"] }, "moodle-availability_date-form": { "requires": ["base", "node", "event", "io", "moodle-core_availability-form"] }, "moodle-availability_grade-form": { "requires": ["base", "node", "event", "moodle-core_availability-form"] }, "moodle-availability_group-form": { "requires": ["base", "node", "event", "moodle-core_availability-form"] }, "moodle-availability_grouping-form": { "requires": ["base", "node", "event", "moodle-core_availability-form"] }, "moodle-availability_profile-form": { "requires": ["base", "node", "event", "moodle-core_availability-form"] }, "moodle-mod_assign-history": { "requires": ["node", "transition"] }, "moodle-mod_quiz-toolboxes": { "requires": ["base", "node", "event", "event-key", "io", "moodle-mod_quiz-quizbase", "moodle-mod_quiz-util-slot", "moodle-core-notification-ajaxexception"] }, "moodle-mod_quiz-autosave": { "requires": ["base", "node", "event", "event-valuechange", "node-event-delegate", "io-form"] }, "moodle-mod_quiz-util": { "requires": ["node", "moodle-core-actionmenu"], "use": ["moodle-mod_quiz-util-base"], "submodules": { "moodle-mod_quiz-util-base": {}, "moodle-mod_quiz-util-slot": { "requires": ["node", "moodle-mod_quiz-util-base"] }, "moodle-mod_quiz-util-page": { "requires": ["node", "moodle-mod_quiz-util-base"] } } }, "moodle-mod_quiz-questionchooser": { "requires": ["moodle-core-chooserdialogue", "moodle-mod_quiz-util", "querystring-parse"] }, "moodle-mod_quiz-dragdrop": { "requires": ["base", "node", "io", "dom", "dd", "dd-scroll", "moodle-core-dragdrop", "moodle-core-notification", "moodle-mod_quiz-quizbase", "moodle-mod_quiz-util-base", "moodle-mod_quiz-util-page", "moodle-mod_quiz-util-slot", "moodle-course-util"] }, "moodle-mod_quiz-modform": { "requires": ["base", "node", "event"] }, "moodle-mod_quiz-quizbase": { "requires": ["base", "node"] }, "moodle-message_airnotifier-toolboxes": { "requires": ["base", "node", "io"] }, "moodle-filter_glossary-autolinker": { "requires": ["base", "node", "io-base", "json-parse", "event-delegate", "overlay", "moodle-core-event", "moodle-core-notification-alert", "moodle-core-notification-exception", "moodle-core-notification-ajaxexception"] }, "moodle-filter_mathjaxloader-loader": { "requires": ["moodle-core-event"] }, "moodle-editor_atto-editor": { "requires": ["node", "transition", "io", "overlay", "escape", "event", "event-simulate", "event-custom", "node-event-html5", "node-event-simulate", "yui-throttle", "moodle-core-notification-dialogue", "moodle-core-notification-confirm", "moodle-editor_atto-rangy", "handlebars", "timers", "querystring-stringify"] }, "moodle-editor_atto-plugin": { "requires": ["node", "base", "escape", "event", "event-outside", "handlebars", "event-custom", "timers", "moodle-editor_atto-menu"] }, "moodle-editor_atto-menu": { "requires": ["moodle-core-notification-dialogue", "node", "event", "event-custom"] }, "moodle-editor_atto-rangy": { "requires": [] }, "moodle-report_eventlist-eventfilter": { "requires": ["base", "event", "node", "node-event-delegate", "datatable", "autocomplete", "autocomplete-filters"] }, "moodle-report_loglive-fetchlogs": { "requires": ["base", "event", "node", "io", "node-event-delegate"] }, "moodle-gradereport_grader-gradereporttable": { "requires": ["base", "node", "event", "handlebars", "overlay", "event-hover"] }, "moodle-gradereport_history-userselector": { "requires": ["escape", "event-delegate", "event-key", "handlebars", "io-base", "json-parse", "moodle-core-notification-dialogue"] }, "moodle-tool_capability-search": { "requires": ["base", "node"] }, "moodle-tool_lp-dragdrop-reorder": { "requires": ["moodle-core-dragdrop"] }, "moodle-tool_monitor-dropdown": { "requires": ["base", "event", "node"] }, "moodle-assignfeedback_editpdf-editor": { "requires": ["base", "event", "node", "io", "graphics", "json", "event-move", "event-resize", "transition", "querystring-stringify-simple", "moodle-core-notification-dialog", "moodle-core-notification-alert", "moodle-core-notification-warning", "moodle-core-notification-exception", "moodle-core-notification-ajaxexception"] }, "moodle-atto_accessibilitychecker-button": { "requires": ["color-base", "moodle-editor_atto-plugin"] }, "moodle-atto_accessibilityhelper-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_align-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_bold-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_charmap-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_clear-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_collapse-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_emojipicker-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_emoticon-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_equation-button": { "requires": ["moodle-editor_atto-plugin", "moodle-core-event", "io", "event-valuechange", "tabview", "array-extras"] }, "moodle-atto_h5p-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_html-button": { "requires": ["promise", "moodle-editor_atto-plugin", "moodle-atto_html-beautify", "moodle-atto_html-codemirror", "event-valuechange"] }, "moodle-atto_html-beautify": {}, "moodle-atto_html-codemirror": { "requires": ["moodle-atto_html-codemirror-skin"] }, "moodle-atto_image-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_indent-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_italic-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_link-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_managefiles-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_managefiles-usedfiles": { "requires": ["node", "escape"] }, "moodle-atto_media-button": { "requires": ["moodle-editor_atto-plugin", "moodle-form-shortforms"] }, "moodle-atto_noautolink-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_orderedlist-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_recordrtc-button": { "requires": ["moodle-editor_atto-plugin", "moodle-atto_recordrtc-recording"] }, "moodle-atto_recordrtc-recording": { "requires": ["moodle-atto_recordrtc-button"] }, "moodle-atto_rtl-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_strike-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_subscript-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_superscript-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_table-button": { "requires": ["moodle-editor_atto-plugin", "moodle-editor_atto-menu", "event", "event-valuechange"] }, "moodle-atto_title-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_underline-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_undo-button": { "requires": ["moodle-editor_atto-plugin"] }, "moodle-atto_unorderedlist-button": { "requires": ["moodle-editor_atto-plugin"] } } }, "gallery": { "name": "gallery", "base": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/yuilib\\/gallery\\/", "combine": true, "comboBase": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/yui_combo.php?", "ext": false, "root": "gallery\\/1633802793\\/", "patterns": { "gallery-": { "group": "gallery" } } } }, "modules": { "core_filepicker": { "name": "core_filepicker", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/repository\\/filepicker.js", "requires": ["base", "node", "node-event-simulate", "json", "async-queue", "io-base", "io-upload-iframe", "io-form", "yui2-treeview", "panel", "cookie", "datatable", "datatable-sort", "resize-plugin", "dd-plugin", "escape", "moodle-core_filepicker", "moodle-core-notification-dialogue"] }, "core_comment": { "name": "core_comment", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/comment\\/comment.js", "requires": ["base", "io-base", "node", "json", "yui2-animation", "overlay", "escape"] }, "mathjax": { "name": "mathjax", "fullpath": "https:\\/\\/cdn.jsdelivr.net\\/npm\\/mathjax@2.7.8\\/MathJax.js?delayStartupUntil=configured" }, "core_question_flags": { "name": "core_question_flags", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/question\\/flags.js", "requires": ["base", "dom", "event-delegate", "io-base"] }, "core_question_engine": { "name": "core_question_engine", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/question\\/qengine.js", "requires": ["node", "event"] }, "mod_quiz": { "name": "mod_quiz", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/mod\\/quiz\\/module.js", "requires": ["base", "dom", "event-delegate", "event-key", "core_question_engine", "moodle-core-formchangechecker"] } } };
        M.yui.loader = { modules: {} };

//]]>
    </script>

    <meta name="viewport" content="width=device-width, initial-scale=1.0">
</head>

<body id="page-mod-quiz-review" class="fpage-mod-quiz-review"
    class="format-topics path-mod path-mod-quiz gecko dir-ltr lang-de yui-skin-sam yui3-skin-sam pagelayout-incourse course-3 context-104 cmid-59 category-1 jsenabled vsc-initialized drawer-ease">

    <div id="page-wrapper" class="d-print-block">

        <script
            src="https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/babel-polyfill/polyfill.min.js"></script>
        <script
            src="https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/polyfills/polyfill.js"></script>
        <script
            src="https://moodle.ruhr-uni-bochum.de/theme/yui_combo.php?rollup/3.17.2/yui-moodlesimple-min.js"></script>
        <script src="https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/javascript-static.js"></script>
        <script>
            document.body.className += ' jsenabled';
        </script>

        <div id="page" class="container-fluid d-print-block">

            <div id="page-content" class="row pb-3 d-print-block">
                <div id="region-main-box" class="col-12">
                    <section id="region-main" class="has-blocks mb-3" aria-label="Inhalt">

                        <div role="main"><span id="maincontent"></span>
                            <form action=""
                                method="post" class="questionflagsaveform">
                                <div>
    '''

moodle_html_footer = '''
  </div>
                            </form>

                        </div>


                    </section>

                </div>
            </div>
        </div>

        <div id="goto-top-link">
            <a class="btn btn-light" role="button" href="#" aria-label="Nach oben">
                <i class="icon fa fa-arrow-up fa-fw " aria-hidden="true"></i>
            </a>
        </div>
        <footer id="page-footer" class="py-3 bg-dark text-light">
            <div class="container">

                <script>
                    //<![CDATA[
                    var require = {
                        baseUrl: 'https://moodle.ruhr-uni-bochum.de/lib/requirejs.php/1633802793/',
                        // We only support AMD modules with an explicit define() statement.
                        enforceDefine: true,
                        skipDataMain: true,
                        waitSeconds: 0,

                        paths: {
                            jquery: 'https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/jquery/jquery-3.5.1.min',
                            jqueryui: 'https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/jquery/ui-1.12.1/jquery-ui.min',
                            jqueryprivate: 'https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/requirejs/jquery-private'
                        },

                        // Custom jquery config map.
                        map: {
                            // '*' means all modules will get 'jqueryprivate'
                            // for their 'jquery' dependency.
                            '*': { jquery: 'jqueryprivate' },
                            // Stub module for 'process'. This is a workaround for a bug in MathJax (see MDL-60458).
                            '*': { process: 'core/first' },

                            // 'jquery-private' wants the real jQuery module
                            // though. If this line was not here, there would
                            // be an unresolvable cyclic dependency.
                            jqueryprivate: { jquery: 'jquery' }
                        }
                    };

//]]>
                </script>
                <script
                    src="https://moodle.ruhr-uni-bochum.de/lib/javascript.php/1633802793/lib/requirejs/require.min.js"></script>
                <script>
                    //<![CDATA[
                    M.util.js_pending("core/first");
                    require(['core/first'], function () {
                        require(['core/prefetch'])
                            ;
                        require(["media_videojs/loader"], function (loader) {
                            loader.setUp('de');
                        });;

                        require(['jquery', 'core/custom_interaction_events'], function ($, CustomEvents) {
                            CustomEvents.define('#single_select6164887fd25b35', [CustomEvents.events.accessibleChange]);
                            $('#single_select6164887fd25b35').on(CustomEvents.events.accessibleChange, function () {
                                var ignore = $(this).find(':selected').attr('data-ignore');
                                if (typeof ignore === typeof undefined) {
                                    $('#single_select_f6164887fd25b34').submit();
                                }
                            });
                        });
                        ;

                        require(['jquery', 'message_popup/notification_popover_controller'], function ($, controller) {
                            var container = $('#nav-notification-popover-container');
                            var controller = new controller(container);
                            controller.registerEventListeners();
                            controller.registerListNavigationEventListeners();
                        });
                        ;

                        require(
                            [
                                'jquery',
                                'core_message/message_popover'
                            ],
                            function (
                                $,
                                Popover
                            ) {
                                var toggle = $('#message-drawer-toggle-6164887fdb8d26164887fd25b38');
                                Popover.init(toggle);
                            });
                        ;

                        require(['jquery', 'core/custom_interaction_events'], function ($, CustomEvents) {
                            CustomEvents.define('#jump-to-activity', [CustomEvents.events.accessibleChange]);
                            $('#jump-to-activity').on(CustomEvents.events.accessibleChange, function () {
                                if (!$(this).val()) {
                                    return false;
                                }
                                $('#url_select_f6164887fd25b317').submit();
                            });
                        });
                        ;

                        require(['jquery', 'core_message/message_drawer'], function ($, MessageDrawer) {
                            var root = $('#message-drawer-6164887fddde16164887fd25b318');
                            MessageDrawer.init(root, '6164887fddde16164887fd25b318', false);
                        });
                        ;

                        require(['jquery', 'core/custom_interaction_events'], function ($, CustomEvents) {
                            CustomEvents.define('#single_select6164887fd25b320', [CustomEvents.events.accessibleChange]);
                            $('#single_select6164887fd25b320').on(CustomEvents.events.accessibleChange, function () {
                                var ignore = $(this).find(':selected').attr('data-ignore');
                                if (typeof ignore === typeof undefined) {
                                    $('#single_select_f6164887fd25b319').submit();
                                }
                            });
                        });
                        ;

                        M.util.js_pending('theme_boost/loader');
                        require(['theme_boost/loader'], function () {
                            M.util.js_complete('theme_boost/loader');
                        });

                        M.util.js_pending('theme_boost/drawer');
                        require(['theme_boost/drawer'], function (drawer) {
                            drawer.init();
                            M.util.js_complete('theme_boost/drawer');
                        });
                        ;
                        M.util.js_pending('qtype_multichoice/answers'); require(['qtype_multichoice/answers'], function (amd) { amd.init("question-8-1"); M.util.js_complete('qtype_multichoice/answers'); });;
                        M.util.js_pending('core/notification'); require(['core/notification'], function (amd) { amd.init(104, []); M.util.js_complete('core/notification'); });;
                        M.util.js_pending('core/log'); require(['core/log'], function (amd) { amd.setConfig({ "level": "warn" }); M.util.js_complete('core/log'); });;
                        M.util.js_pending('core/page_global'); require(['core/page_global'], function (amd) { amd.init(); M.util.js_complete('core/page_global'); });
                        M.util.js_complete("core/first");
                    });
//]]>
                </script>
                <script>
                    //<![CDATA[
                    M.yui.add_module({ "qtype_multianswer": { "name": "qtype_multianswer", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/question\\/type\\/multianswer\\/module.js", "requires": ["base", "node", "event", "overlay"] }, "core_question_engine": { "name": "core_question_engine", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/question\\/qengine.js", "requires": ["node", "event"] }, "mod_quiz": { "name": "mod_quiz", "fullpath": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/lib\\/javascript.php\\/1633802793\\/mod\\/quiz\\/module.js", "requires": ["base", "dom", "event-delegate", "event-key", "core_question_engine", "moodle-core-formchangechecker"] } });

//]]>
                </script>
                <script>
                    //<![CDATA[
                    M.str = { "moodle": { "lastmodified": "Zuletzt ge\\u00e4ndert", "name": "Name", "error": "Fehler", "info": "Infos", "yes": "Ja", "no": "Nein", "cancel": "Abbrechen", "changesmadereallygoaway": "Sie haben etwas ver\\u00e4ndert. M\\u00f6chten Sie die Seite wirklich verlassen, ohne die \\u00c4nderungen zu speichern?", "confirm": "Best\\u00e4tigen", "areyousure": "Sind Sie sicher?", "closebuttontitle": "Schlie\\u00dfen", "unknownerror": "Unbekannter Fehler", "file": "Datei", "url": "URL" }, "repository": { "type": "Typ", "size": "Gr\\u00f6\\u00dfe", "invalidjson": "Ung\\u00fcltiger JSON-Text", "nofilesattached": "Keine Datei", "filepicker": "Dateiauswahl", "logout": "Abmelden", "nofilesavailable": "Keine Dateien vorhanden", "norepositoriesavailable": "Sie k\\u00f6nnen hier zur Zeit keine Dateien hochladen.", "fileexistsdialogheader": "Datei bereits vorhanden", "fileexistsdialog_editor": "Eine Datei mit diesem Namen wurde bereits an den Text angeh\\u00e4ngt, den Sie gerade bearbeiten", "fileexistsdialog_filemanager": "Eine Datei mit diesem Namen wurde bereits an den Text angeh\\u00e4ngt", "renameto": "Nach '{$a}' umbenennen", "referencesexist": "Es gibt {$a} Links zu dieser Datei.", "select": "W\\u00e4hlen Sie" }, "admin": { "confirmdeletecomments": "M\\u00f6chten Sie die Kommentare wirklich l\\u00f6schen?", "confirmation": "Best\\u00e4tigung" }, "question": { "flagged": "markiert" }, "quiz": { "functiondisabledbysecuremode": "Diese Funktion ist aktuell deaktiviert", "startattempt": "Versuch beginnen", "timesup": "Zeit ist abgelaufen." }, "debug": { "debuginfo": "Debug-Info", "line": "Zeile", "stacktrace": "Stack trace" }, "langconfig": { "labelsep": ":\\u00a0" } };
//]]>
                </script>
                <script>
                    //<![CDATA[
                    (function () {
                        Y.use("moodle-filter_mathjaxloader-loader", function () {
                            M.filter_mathjaxloader.configure({ "mathjaxconfig": "\\nMathJax.Hub.Config({\\n    config: [\\"Accessible.js\\", \\"Safe.js\\"],\\n    errorSettings: { message: [\\"!\\"] },\\n    skipStartupTypeset: true,\\n    messageStyle: \\"none\\"\\n});\\n", "lang": "de" });
                        });
                        M.util.js_pending('random6164887fd25b31'); Y.use('core_question_flags', function (Y) { M.core_question_flags.init(Y, "https:\\/\\/moodle.ruhr-uni-bochum.de\\/question\\/toggleflag.php", [{ "src": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/image.php\\/boost\\/core\\/1633802793\\/i\\/unflagged", "title": "Diese Frage als Referenz markieren", "alt": "nicht markiert" }, { "src": "https:\\/\\/moodle.ruhr-uni-bochum.de\\/theme\\/image.php\\/boost\\/core\\/1633802793\\/i\\/flagged", "title": "Markierung entfernen", "alt": "markiert" }], ["Frage markieren", "Markierung entfernen"]); M.util.js_complete('random6164887fd25b31'); });
                        M.util.js_pending('random6164887fd25b32'); Y.use('mod_quiz', function (Y) { M.mod_quiz.nav.init(Y); M.util.js_complete('random6164887fd25b32'); });
                        M.util.help_popups.setup(Y);
                        M.util.js_pending('random6164887fd25b321'); Y.use('qtype_multianswer', function (Y) { M.qtype_multianswer.init(Y, "#question-8-5"); M.util.js_complete('random6164887fd25b321'); });
                        M.util.js_pending('random6164887fd25b322'); Y.use('mod_quiz', function (Y) { M.mod_quiz.init_review_form(Y); M.util.js_complete('random6164887fd25b322'); });
                        M.util.js_pending('random6164887fd25b323'); Y.on('domready', function () { M.util.js_complete("init"); M.util.js_complete('random6164887fd25b323'); });
                    })();
//]]>
                </script>

            </div>
        </footer>
    </div>

</body>

</html>
    '''



import os

from pathlib import Path

from moodle_dl.utils import cutie
from moodle_dl.state_recorder.file import File
from moodle_dl.moodle_connector.moodle_service import MoodleService
from moodle_dl.config_service.config_helper import ConfigHelper
from moodle_dl.state_recorder.state_recorder import StateRecorder


class OfflineService:
    def __init__(self, config_helper: ConfigHelper, storage_path: str):
        self.config_helper = config_helper
        self.storage_path = storage_path
        self.state_recorder = StateRecorder(Path(storage_path) / 'moodle_state.db')

    def interactively_manage_database(self):
        RESET_SEQ = '\033[0m'
        COLOR_SEQ = '\033[1;%dm'

        BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)

        stored_files = self.state_recorder.get_stored_files()

        stored_files = MoodleService.filter_courses(stored_files, self.config_helper)

        if len(stored_files) <= 0:
            return

        print(
            'This management tool will navigate you through a menu to'
            + ' selectively remove file entries from the database so'
            + ' that these files can be downloaded again.'
        )

        course_options = []
        courses = []
        for course in stored_files:
            for course_file in course.files:
                if not os.path.exists(course_file.saved_to):
                    course_options.append(COLOR_SEQ % BLUE + course.fullname + RESET_SEQ)
                    courses.append(course)
                    break

        print('Choose one of the courses:')
        print('[Confirm your selection with the Enter key]')
        print('')
        selected_course_id = cutie.select(options=course_options)

        selected_course = courses[selected_course_id]

        section_options = []
        sections = []

        # Add the option to select all sections
        section_options.append(COLOR_SEQ % MAGENTA + '[All sections]' + RESET_SEQ)
        sections.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if not os.path.exists(course_file.saved_to) and (course_file.section_name not in sections):
                section_options.append(COLOR_SEQ % MAGENTA + course_file.section_name + RESET_SEQ)
                sections.append(course_file.section_name)

        print('From which sections you want to select files.')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')

        selected_sections_ids = cutie.select_multiple(options=section_options, minimal_count=1)

        selected_sections = []
        for selected_sections_id in selected_sections_ids:
            if selected_sections_id == 0:
                selected_sections = sections[1:]
                break
            elif (selected_sections_id) < len(sections):
                selected_sections.append(sections[selected_sections_id])

        file_options = []
        files = []

        # Add the option to select all files
        file_options.append(COLOR_SEQ % CYAN + '[All files]' + RESET_SEQ)
        files.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if not os.path.exists(course_file.saved_to) and (course_file.section_name in selected_sections):
                file_options.append(COLOR_SEQ % CYAN + course_file.content_filename + RESET_SEQ)
                files.append(course_file)

        print('Which of the files should be removed form the database, so that they will be redownloaded?')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')
        selected_files = cutie.select_multiple(options=file_options)

        files_to_delete = []
        for file_index in selected_files:
            if file_index == 0:  # If all files is selected
                for file_to_delete in files[1:]:  # Ignore the first element of the array set as None
                    if isinstance(file_to_delete, File):
                        files_to_delete.append(file_to_delete)

                break

            elif file_index < len(files) and isinstance(files[file_index], File):
                files_to_delete.append(files[file_index])

        self.state_recorder.batch_delete_files_from_db(files_to_delete)



#!/usr/bin/env python3
# coding=utf-8

import os
import sys
import logging
import argparse
import traceback

from shutil import which

import sentry_sdk

import moodle_dl.utils.process_lock as process_lock


try:
    # In unix readline needs to be loaded so that
    # arrowkeys work in input
    import readline  # pylint: disable=unused-import
except ImportError:
    pass

from logging.handlers import RotatingFileHandler
from moodle_dl.utils import cutie
from moodle_dl.utils.logger import Log
from moodle_dl.version import __version__
from moodle_dl.download_service.path_tools import PathTools
from moodle_dl.config_service.config_helper import ConfigHelper
from moodle_dl.config_service.config_service import ConfigService
from moodle_dl.state_recorder.offline_service import OfflineService
from moodle_dl.moodle_connector.moodle_service import MoodleService
from moodle_dl.download_service.download_service import DownloadService
from moodle_dl.notification_services.mail.mail_service import MailService
from moodle_dl.notification_services.xmpp.xmpp_service import XmppService
from moodle_dl.download_service.fake_download_service import FakeDownloadService
from moodle_dl.notification_services.console.console_service import ConsoleService
from moodle_dl.notification_services.telegram.telegram_service import TelegramService

IS_DEBUG = False


class ReRaiseOnError(logging.StreamHandler):
    """
    A logging-handler class which allows the exception-catcher of i.e. PyCharm
    to intervine
    """

    def emit(self, record):
        if hasattr(record, 'exception'):
            raise record.exception


def run_init(storage_path, use_sso=False, skip_cert_verify=False):
    config = ConfigHelper(storage_path)

    if config.is_present():
        do_override_input = cutie.prompt_yes_or_no(Log.error_str('Do you want to override the existing config?'))

        if not do_override_input:
            sys.exit(0)

    MailService(config).interactively_configure()
    TelegramService(config).interactively_configure()
    XmppService(config).interactively_configure()

    do_sentry = cutie.prompt_yes_or_no('Do you want to configure Error Reporting via Sentry?')
    if do_sentry:
        sentry_dsn = input('Please enter your Sentry DSN:   ')
        config.set_property('sentry_dsn', sentry_dsn)

    moodle = MoodleService(config, storage_path, skip_cert_verify)

    if use_sso:
        moodle.interactively_acquire_sso_token()
    else:
        moodle.interactively_acquire_token()

    Log.success('Configuration finished and saved!')

    if os.name != 'nt':
        if storage_path == '.':
            Log.info(
                '  To set a cron-job for this program on your Unix-System:\n'
                + '    1. `crontab -e`\n'
                + '    2. Add `*/15 * * * * cd "{}" && moodle-dl`\n'.format(os.getcwd())
                + '    3. Save and you\'re done!'
            )
        else:
            Log.info(
                '  To set a cron-job for this program on your Unix-System:\n'
                + '    1. `crontab -e`\n'
                + '    2. Add `*/15 * * * * cd "{}" && moodle-dl -p "{}"`\n'.format(os.getcwd(), storage_path)
                + '    3. Save and you\'re done!'
            )

    print('')

    Log.info('You can always do the additional configuration later with the --config option.')

    do_config = cutie.prompt_yes_or_no('Do you want to make additional configurations now?')

    if do_config:
        run_configure(storage_path, skip_cert_verify)

    print('')
    Log.success('All set and ready to go!')


def run_configure(storage_path, skip_cert_verify=False):
    config = ConfigHelper(storage_path)
    config.load()  # because we do not want to override the other settings

    ConfigService(config, storage_path, skip_cert_verify).interactively_acquire_config()

    Log.success('Configuration successfully updated!')


def run_new_token(storage_path, use_sso=False, username: str = None, password: str = None, skip_cert_verify=False):
    config = ConfigHelper(storage_path)
    config.load()  # because we do not want to override the other settings

    moodle = MoodleService(config, storage_path, skip_cert_verify)

    if use_sso:
        moodle.interactively_acquire_sso_token(use_stored_url=True)
    else:
        moodle.interactively_acquire_token(use_stored_url=True, username=username, password=password)

    Log.success('New Token successfully saved!')


def run_manage_database(storage_path):
    config = ConfigHelper(storage_path)
    config.load()  # because we want to only manage configured courses

    offline_service = OfflineService(config, storage_path)
    offline_service.interactively_manage_database()

    Log.success('All done.')


def run_change_notification_mail(storage_path):
    config = ConfigHelper(storage_path)
    config.load()

    MailService(config).interactively_configure()

    Log.success('Configuration successfully updated!')


def run_change_notification_telegram(storage_path):
    config = ConfigHelper(storage_path)
    config.load()

    TelegramService(config).interactively_configure()

    Log.success('Telegram Configuration successfully updated!')


def run_change_notification_xmpp(storage_path):
    config = ConfigHelper(storage_path)
    config.load()

    XmppService(config).interactively_configure()

    Log.success('XMPP Configuration successfully updated!')


def run_main(
    storage_path,
    verbose=False,
    skip_cert_verify=False,
    ignore_ytdl_errors=False,
    without_downloading_files=False,
    log_responses=False,
):

    log_formatter = logging.Formatter('%(asctime)s  %(levelname)s  {%(module)s}  %(message)s', '%Y-%m-%d %H:%M:%S')
    log_file = os.path.join(storage_path, 'MoodleDownloader.log')
    log_handler = RotatingFileHandler(
        log_file, mode='a', maxBytes=1 * 1024 * 1024, backupCount=2, encoding='utf-8', delay=0
    )

    log_handler.setFormatter(log_formatter)
    if verbose:
        log_handler.setLevel(logging.DEBUG)
    else:
        log_handler.setLevel(logging.INFO)

    app_log = logging.getLogger()
    if verbose:
        app_log.setLevel(logging.DEBUG)
    else:
        app_log.setLevel(logging.INFO)
    app_log.addHandler(log_handler)

    logging.info('--- moodle-dl started ---------------------')
    Log.info('Moodle Downloader starting...')
    if verbose:
        logging.debug('moodle-dl version: %s', __version__)
        logging.debug('python version: %s', ".".join(map(str, sys.version_info[:3])))
        ffmpeg_available = which('ffmpeg') is not None
        logging.debug('Is ffmpeg available: %s', ffmpeg_available)

    if IS_DEBUG:
        logging.info('Debug-Mode detected. Errors will be re-risen.')
        app_log.addHandler(ReRaiseOnError())

    try:
        msg_load_config = 'Loading config...'
        logging.debug(msg_load_config)
        Log.debug(msg_load_config)

        config = ConfigHelper(storage_path)
        config.load()
    except BaseException as e:
        logging.error('Error while trying to load the Configuration! %s Exiting...', e, extra={'exception': e})
        Log.error('Error while trying to load the Configuration!')
        sys.exit(1)

    r_client = False
    try:
        sentry_dsn = config.get_property('sentry_dsn')
        if sentry_dsn:
            sentry_sdk.init(sentry_dsn)
    except BaseException:
        pass

    mail_service = MailService(config)
    tg_service = TelegramService(config)
    xmpp_service = XmppService(config)
    console_service = ConsoleService(config)

    PathTools.restricted_filenames = config.get_restricted_filenames()

    try:
        if not IS_DEBUG:
            process_lock.lock(storage_path)

        moodle = MoodleService(config, storage_path, skip_cert_verify, log_responses)

        msg_checking_for_changes = 'Checking for changes for the configured Moodle-Account....'
        logging.debug(msg_checking_for_changes)
        Log.debug(msg_checking_for_changes)
        changed_courses = moodle.fetch_state()

        if log_responses:
            msg_responses_logged = (
                "All JSON-responses from Moodle have been written to the responses.log file. Exiting..."
            )
            logging.debug(msg_responses_logged)
            Log.success(msg_responses_logged)
            process_lock.unlock(storage_path)
            return

        msg_start_downloading = 'Start downloading changed files...'
        logging.debug(msg_start_downloading)
        Log.debug(msg_start_downloading)

        if without_downloading_files:
            downloader = FakeDownloadService(changed_courses, moodle, storage_path)
        else:
            downloader = DownloadService(changed_courses, moodle, storage_path, skip_cert_verify, ignore_ytdl_errors)
        downloader.run()
        failed_downloads = downloader.get_failed_url_targets()

        changed_courses_to_notify = moodle.recorder.changes_to_notify()

        if len(changed_courses_to_notify) > 0:
            console_service.notify_about_changes_in_moodle(changed_courses_to_notify)
            mail_service.notify_about_changes_in_moodle(changed_courses_to_notify)
            tg_service.notify_about_changes_in_moodle(changed_courses_to_notify)
            xmpp_service.notify_about_changes_in_moodle(changed_courses_to_notify)

            moodle.recorder.notified(changed_courses_to_notify)

        else:
            msg_no_changes = 'No changes found for the configured Moodle-Account.'
            logging.info(msg_no_changes)
            Log.warning(msg_no_changes)

        if len(failed_downloads) > 0:
            console_service.notify_about_failed_downloads(failed_downloads)
            mail_service.notify_about_failed_downloads(failed_downloads)
            tg_service.notify_about_failed_downloads(failed_downloads)
            xmpp_service.notify_about_failed_downloads(failed_downloads)

        process_lock.unlock(storage_path)

        logging.debug('All done. Exiting...')
        Log.success('All done. Exiting..')
    except BaseException as e:
        print('\n')
        if not isinstance(e, process_lock.LockError):
            process_lock.unlock(storage_path)

        error_formatted = traceback.format_exc()
        logging.error(error_formatted, extra={'exception': e})

        if r_client:
            sentry_sdk.capture_exception(e)

        if verbose:
            Log.critical('Exception:\n%s' % (error_formatted))

        short_error = str(e)
        if not short_error or short_error.isspace():
            short_error = traceback.format_exc(limit=1)

        console_service.notify_about_error(short_error)
        mail_service.notify_about_error(short_error)
        tg_service.notify_about_error(short_error)
        xmpp_service.notify_about_error(short_error)

        logging.debug('Exception-Handling completed. Exiting...')

        sys.exit(1)


def _dir_path(path):
    # Working around MAX_PATH limitation on Windows (see
    # http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx)
    if os.name == 'nt':
        absfilepath = os.path.abspath(path)
        path = '\\\\?\\' + absfilepath

    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError('"%s" is not a valid path. Make sure the directory exists.' % (str(path)))


def check_debug():
    global IS_DEBUG
    if 'pydevd' in sys.modules:
        IS_DEBUG = True
        Log.debug('[RUNNING IN DEBUG-MODE!]')


def get_parser():
    """
    Creates a new argument parser.
    """
    parser = argparse.ArgumentParser(
        description=('Moodle Downloader 2 helps you download all the course files  of your Moodle account.')
    )
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--version', action='version', version='moodle-dl ' + __version__, help='Print program version and exit'
    )

    group.add_argument(
        '-i',
        '--init',
        action='store_true',
        help=(
            'Guides you trough the configuration of the'
            + ' software, including the activation of'
            + ' mail-notifications and obtainment of a'
            + ' login-token for your Moodle-Account. It'
            + ' does not fetch the current state of your'
            + ' Moodle-Account.'
        ),
    )

    group.add_argument(
        '-c',
        '--config',
        action='store_true',
        help=(
            'Guides you through the additional'
            + ' configuration of the software. This'
            + ' includes the selection of the courses to'
            + ' be downloaded and various configuration'
            + ' options for these courses.'
        ),
    )

    group.add_argument(
        '-nt',
        '--new-token',
        action='store_true',
        help=(
            'Overrides the login-token with a newly obtained'
            + ' one. It does not fetch the current state of your'
            + ' Moodle-Account. Use it if at any point in time,'
            + ' for whatever reason, the saved token gets'
            + ' rejected by Moodle. It does not affect the rest'
            + ' of the config.'
        ),
    )

    group.add_argument(
        '-cm',
        '--change-notification-mail',
        action='store_true',
        help=(
            'Activate/deactivate/change the settings for'
            + ' receiving notifications via e-mail. It does not'
            + ' affect the rest of the config.'
        ),
    )

    group.add_argument(
        '-ct',
        '--change-notification-telegram',
        action='store_true',
        help=(
            'Activate/deactivate/change the settings for'
            + ' receiving notifications via Telegram. It does not'
            + ' affect the rest of the config.'
        ),
    )

    group.add_argument(
        '-cx',
        '--change-notification-xmpp',
        action='store_true',
        help=(
            'Activate/deactivate/change the settings for'
            + ' receiving notifications via XMPP. It does not'
            + ' affect the rest of the config.'
        ),
    )

    group.add_argument(
        '-md',
        '--manage-database',
        action='store_true',
        help=(
            'This option lets you manage the offline database.'
            + ' It allows you to delete entries from the database'
            + ' that are no longer available locally so that they'
            + ' can be downloaded again.'
        ),
    )

    group.add_argument(
        '--log-responses',
        default=False,
        action='store_true',
        help='To generate a responses.log file'
        + ' in which all JSON responses from Moodles are logged'
        + ' along with the requested URL.',
    )

    parser.add_argument(
        '-p',
        '--path',
        default='.',
        type=_dir_path,
        help=(
            'Sets the location of the configuration,'
            + ' logs and downloaded files. PATH must be an'
            + ' existing directory in which you have read and'
            + ' write access. (default: current working'
            + ' directory)'
        ),
    )

    parser.add_argument(
        '-t',
        '--threads',
        default=5,
        type=int,
        help=('Sets the number of download threads. (default: %(default)s)'),
    )

    parser.add_argument(
        '-u',
        '--username',
        default=None,
        type=str,
        help=('Specify username to skip the query when creating a new token.'),
    )

    parser.add_argument(
        '-pw',
        '--password',
        default=None,
        type=str,
        help=('Specify password to skip the query when creating a new token.'),
    )

    parser.add_argument(
        '-v',
        '--verbose',
        default=False,
        action='store_true',
        help='Print various debugging information',
    )

    parser.add_argument(
        '--skip-cert-verify',
        default=False,
        action='store_true',
        help='If this flag is set, the SSL certificate '
        + 'is not verified. This option should only be used in '
        + 'non production environments.',
    )

    parser.add_argument(
        '-iye',
        '--ignore-ytdl-errors',
        default=False,
        action='store_true',
        help='If this option is set, errors that occur when downloading with the help of Youtube-dl are ignored. '
        + 'Thus, no further attempt will be made to download the file using youtube-dl. '
        + 'By default, youtube-dl errors are critical, so the download of the corresponding file '
        + 'will be aborted and when you run moodle-dl again, the download will be repeated.',
    )

    parser.add_argument(
        '--without-downloading-files',
        default=False,
        action='store_true',
        help='If this flag is set, no files are downloaded.'
        + ' This allows the local database to be updated without'
        + ' having to download all files.',
    )

    parser.add_argument(
        '-sso',
        '--sso',
        default=False,
        action='store_true',
        help='This flag can be used together with --init and -nt. If'
        + ' this flag is set, you will be guided through the'
        + ' Single Sign On (SSO) login process during'
        + ' initialization or new token retrieval.',
    )

    return parser


# --- called at the program invocation: -------------------------------------
def main(args=None):
    """The main routine."""

    check_debug()

    parser = get_parser()
    args = parser.parse_args(args)

    use_sso = args.sso
    verbose = args.verbose
    username = args.username
    password = args.password
    storage_path = args.path
    skip_cert_verify = args.skip_cert_verify
    ignore_ytdl_errors = args.ignore_ytdl_errors
    without_downloading_files = args.without_downloading_files
    log_responses = args.log_responses
    DownloadService.thread_count = args.threads

    if args.init:
        run_init(storage_path, use_sso, skip_cert_verify)
    elif args.config:
        run_configure(storage_path, skip_cert_verify)
    elif args.new_token:
        run_new_token(storage_path, use_sso, username, password, skip_cert_verify)
    elif args.change_notification_mail:
        run_change_notification_mail(storage_path)
    elif args.change_notification_telegram:
        run_change_notification_telegram(storage_path)
    elif args.change_notification_xmpp:
        run_change_notification_xmpp(storage_path)
    elif args.manage_database:
        run_manage_database(storage_path)
    else:
        run_main(storage_path, verbose, skip_cert_verify, ignore_ytdl_errors, without_downloading_files, log_responses)



from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.url_target import URLTarget


class TelegramFormater:
    @staticmethod
    def append_with_limit(new_line: str, one_msg_content: str, msg_list: [str]):
        """Appends a new line to a message string,
        if the string is to long it ist appended to the message list.
        Returns the new message string.

        Args:
            new_line (str): A new line to append to a message string
            one_msg_content (str): The current message string
            msg_list ([str]): The list of finished messages
        Returns:
            str: The new message
        """
        # Replace Telegram Entities
        new_line = new_line.replace('<', '&lt;').replace('>', '&gt')
        if len(one_msg_content) + len(new_line) >= 4096:
            msg_list.append(one_msg_content)
            return new_line
        else:
            return one_msg_content + new_line

    @classmethod
    def create_full_moodle_diff_messages(cls, changed_courses: [Course]) -> [str]:
        """
        Creates telegram messages with all changed files. This includes new,
        modified and deleted files. Files that have changed since the last message.

        @param changed_courses: A list of all courses with their modified files.
        @returns a list of messages
        """

        diff_count = 0
        for course in changed_courses:
            diff_count += len(course.files)

        result_list = []
        one_msg_content = '%s new Changes in the Moodle courses!' % (diff_count)

        for course in changed_courses:
            new_line = '\r\n\r\n\r\n👉 **' + course.fullname + '**\r\n'
            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

            for file in course.files:
                saved_to_path = file.saved_to
                if file.new_file is not None:
                    saved_to_path = file.new_file.saved_to
                if file.modified:
                    new_line = '\r\n🔄 Modified:\r\n ' + saved_to_path
                elif file.moved:
                    new_line = '\r\n🔀 Moved:\r\n ' + saved_to_path
                elif file.deleted:
                    new_line = '\r\n❌ Deleted:\r\n ' + saved_to_path
                else:
                    new_line = '\r\n✅ Added:\r\n ' + saved_to_path

                one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

                if file.content_type == 'description':
                    try:
                        with open(saved_to_path, 'r') as description_file:
                            description_lines = description_file.read().splitlines()
                    except Exception:
                        description_lines = []

                    if len(description_lines) > 0:
                        one_msg_content = cls.append_with_limit('\r\n👇\r\n', one_msg_content, result_list)
                        for new_line in description_lines:
                            new_line = new_line + '\r\n'
                            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)
                        one_msg_content = cls.append_with_limit('👆\r\n', one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list

    @classmethod
    def create_full_error_messages(cls, details) -> [str]:
        """
        Creates error messages
        """
        result_list = []

        one_msg_content = '🛑 The following error occurred during execution:\r\n'
        for new_line in details.splitlines():
            new_line = new_line + '\r\n'
            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list

    @classmethod
    def create_full_failed_downloads_messages(cls, failed_downloads: [URLTarget]) -> [str]:
        """
        Creates messages with all failed downloads
        """

        result_list = []
        if len(failed_downloads) == 0:
            return result_list

        one_msg_content = '⁉️ Error while trying to download files, look at the log for more details.\r\nList of failed downloads:\r\n\r\n'
        for url_target in failed_downloads:
            new_line = '⚠️ %s:\r\n%s\r\n\r\n' % (url_target.file.content_filename, url_target.error)
            if url_target.file.content_filename != url_target.file.content_fileurl:
                new_line = '⚠️ %s (%s):\r\n%s\r\n\r\n' % (
                    url_target.file.content_filename,
                    url_target.file.content_fileurl,
                    url_target.error,
                )

            one_msg_content = cls.append_with_limit(new_line, one_msg_content, result_list)

        result_list.append(one_msg_content)
        return result_list



RESET_SEQ = '\033[0m'
COLOR_SEQ = '\033[1;%dm'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)


class Log:
    """
    Logs a given string to output with colors
    :param logString: the string that should be logged

    The string functions returns the strings that would be logged.
    """

    @staticmethod
    def info_str(logString: str):
        return COLOR_SEQ % WHITE + logString + RESET_SEQ

    @staticmethod
    def special_str(logString: str):
        return COLOR_SEQ % BLUE + logString + RESET_SEQ

    @staticmethod
    def debug_str(logString: str):
        return COLOR_SEQ % CYAN + logString + RESET_SEQ

    @staticmethod
    def warning_str(logString: str):
        return COLOR_SEQ % YELLOW + logString + RESET_SEQ

    @staticmethod
    def error_str(logString: str):
        return COLOR_SEQ % RED + logString + RESET_SEQ

    @staticmethod
    def critical_str(logString: str):
        return COLOR_SEQ % MAGENTA + logString + RESET_SEQ

    @staticmethod
    def success_str(logString: str):
        return COLOR_SEQ % GREEN + logString + RESET_SEQ

    @staticmethod
    def info(logString: str):
        print(Log.info_str(logString))

    @staticmethod
    def special(logString: str):
        print(Log.special_str(logString))

    @staticmethod
    def debug(logString: str):
        print(Log.debug_str(logString))

    @staticmethod
    def warning(logString: str):
        print(Log.warning_str(logString))

    @staticmethod
    def error(logString: str):
        print(Log.error_str(logString))

    @staticmethod
    def critical(logString: str):
        print(Log.critical_str(logString))

    @staticmethod
    def success(logString: str):
        print(Log.success_str(logString))



import logging
import traceback

from getpass import getpass

from moodle_dl.utils import cutie
from moodle_dl.utils.logger import Log
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.url_target import URLTarget
from moodle_dl.notification_services.mail.mail_shooter import MailShooter
from moodle_dl.notification_services.notification_service import NotificationService
from moodle_dl.notification_services.mail.mail_formater import (
    create_full_welcome_mail,
    create_full_moodle_diff_mail,
    create_full_error_mail,
    create_full_failed_downloads_mail,
)


class MailService(NotificationService):
    def interactively_configure(self) -> None:
        """
        Guides the user through the configuration of the mail notification.
        """

        do_mail = cutie.prompt_yes_or_no('Do you want to activate Notifications via mail?')

        if not do_mail:
            self.config_helper.remove_property('mail')
        else:
            print('[The following Inputs are not validated!]')

            config_valid = False
            while not config_valid:
                sender = input('E-Mail-Address of the Sender:   ')
                server_host = input('Host of the SMTP-Server:   ')
                server_port = input('Port of the SMTP-Server [STARTTLS, default 587]:   ')
                if server_port == '':
                    print('Using default port 587!')
                    server_port = '587'
                username = input('Username for the SMTP-Server:   ')
                password = getpass('Password for the SMTP-Server [no output]:   ')
                target = input('E-Mail-Address of the Target:   ')

                print('Testing Mail-Config...')
                welcome_content = create_full_welcome_mail()
                mail_shooter = MailShooter(sender, server_host, int(server_port), username, password)
                try:
                    mail_shooter.send(target, 'Hey!', welcome_content[0], welcome_content[1])
                except BaseException as e:
                    print('Error while sending the test mail: %s' % (str(e)))
                    continue
                else:
                    input(
                        'Please check if you received the Welcome-Mail.'
                        + ' If yes, confirm with Return.\nIf not, exit'
                        + ' this program ([CTRL]+[C]) and try again later.'
                    )
                    config_valid = True

                raw_send_error_msg = ''
                while raw_send_error_msg not in ['y', 'n']:
                    raw_send_error_msg = input('Do you want to also get error reports sent by mail? [y/n]   ')
                do_send_error_msg = raw_send_error_msg == 'y'

                mail_cfg = {
                    'sender': sender,
                    'server_host': server_host,
                    'server_port': server_port,
                    'username': username,
                    'password': password,
                    'target': target,
                    'send_error_msg': do_send_error_msg,
                }

                self.config_helper.set_property('mail', mail_cfg)

    def _is_configured(self) -> bool:
        # Checks if the sending of emails has been configured.
        try:
            self.config_helper.get_property('mail')
            return True
        except ValueError:
            logging.debug('Mail-Notifications not configured, skipping.')
            return False

    def _send_mail(self, subject, mail_content: (str, {str: str})):
        """
        Sends an email
        """
        if not self._is_configured():
            return

        mail_cfg = self.config_helper.get_property('mail')

        try:
            logging.debug('Sending Notification via Mail...')
            Log.debug('Sending Notification via Mail... (Please wait)')

            mail_shooter = MailShooter(
                mail_cfg['sender'],
                mail_cfg['server_host'],
                int(mail_cfg['server_port']),
                mail_cfg['username'],
                mail_cfg['password'],
            )
            mail_shooter.send(mail_cfg['target'], subject, mail_content[0], mail_content[1])
        except BaseException as e:
            error_formatted = traceback.format_exc()
            logging.error('While sending notification:\n%s', error_formatted, extra={'exception': e})
            raise e  # to be properly notified via Sentry

    def notify_about_changes_in_moodle(self, changes: [Course]) -> None:
        """
        Sends out a notification email about the downloaded changes.
        @param changes: A list of changed courses with changed files.
        """
        if not self._is_configured():
            return

        mail_content = create_full_moodle_diff_mail(changes)

        diff_count = 0
        for course in changes:
            diff_count += len(course.files)

        self._send_mail('%s new Changes in the Moodle courses!' % (diff_count), mail_content)

    def notify_about_error(self, error_description: str):
        """
        Sends out an error mail if configured to do so.
        @param error_description: The error object.
        """
        if not self._is_configured():
            return

        mail_cfg = self.config_helper.get_property('mail')

        if not mail_cfg.get('send_error_msg', True):
            return

        mail_content = create_full_error_mail(error_description)
        self._send_mail('Error!', mail_content)

    def notify_about_failed_downloads(self, failed_downloads: [URLTarget]):
        """
        Sends out an mail with all failed download if configured to send out error messages.
        @param failed_downloads: A list of failed URLTargets.
        """
        if not self._is_configured():
            return

        mail_cfg = self.config_helper.get_property('mail')

        if not mail_cfg.get('send_error_msg', True):
            return

        mail_content = create_full_failed_downloads_mail(failed_downloads)
        self._send_mail('Faild to download files!', mail_content)



import os
import platform

from pathlib import Path

from moodle_dl.utils.logger import Log
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.path_tools import PathTools
from moodle_dl.moodle_connector.moodle_service import MoodleService
from moodle_dl.download_service.download_service import DownloadService


class FakeDownloadService:
    """
    FakeDownloadService fakes a DownloadService.
    This way a local database of Moodle's current files
    can be created without actually downloading the files.
    """

    def __init__(self, courses: [Course], moodle_service: MoodleService, storage_path: str):
        """
        Initiates the FakeDownloadService with all files that
        need to be downloaded (saved in the database).
        @param courses: A list of courses that contains all modified files.
        @param moodle_service: A reference to the moodle_service, currently
                               only to get to the state_recorder.
        @param storage_path: The location where the files would be saved.
        """

        self.courses = courses
        self.state_recorder = moodle_service.recorder
        self.storage_path = storage_path

        # delete files, that should be deleted
        self.state_recorder.batch_delete_files(self.courses)

        # Prepopulate queue with any files that were given
        for course in self.courses:
            for file in course.files:
                if file.deleted is False:

                    save_destination = DownloadService.gen_path(self.storage_path, course, file)

                    filename = PathTools.to_valid_name(file.content_filename)

                    file.saved_to = str(Path(save_destination) / filename)

                    if file.module_modname.startswith('url'):
                        file.saved_to = str(Path(save_destination) / (filename + '.desktop'))
                        if os.name == 'nt' or platform.system() == "Darwin":
                            file.saved_to = str(Path(save_destination) / (filename + '.URL'))

                    if file.content_type == 'description':
                        file.saved_to = str(Path(save_destination) / (filename + '.md'))

                    self.state_recorder.save_file(file, course.id, course.fullname)

    def get_failed_url_targets(self):
        """
        Return a list of failed Downloads, as a list of URLTargets.
        No download can fail, so this is only a dummy function.
        """
        return []

    def run(self):
        """Dummy function"""
        Log.success('All files stored in the Database!')



from moodle_dl.state_recorder.course import Course
from moodle_dl.moodle_connector.request_helper import RequestHelper


class AssignmentsHandler:
    """
    Fetches and parses the various endpoints in Moodle.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_assignments(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Assignments List for all courses from the
        Moodle system
        @return: A Dictionary of all assignments,
                 indexed by courses, then assignment
        """
        # do this only if version is greater then 2.4
        # because mod_assign_get_assignments will fail
        if self.version < 2012120300:
            return {}

        print('\rDownloading assignments information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        assign_result = self.request_helper.post_REST('mod_assign_get_assignments', extra_data)

        assign_courses = assign_result.get('courses', [])

        result = {}
        for assign_course in assign_courses:
            course_id = assign_course.get('id', 0)
            course_assigns = {}
            course_assign_objs = assign_course.get('assignments', [])

            for course_assign_obj in course_assign_objs:
                assign_id = course_assign_obj.get('cmid', 0)
                assign_rid = course_assign_obj.get('id', 0)
                assign_name = course_assign_obj.get('name', '')
                assign_timemodified = course_assign_obj.get('timemodified', 0)

                assign_files = []
                assign_files += course_assign_obj.get('introfiles', [])
                assign_files += course_assign_obj.get('introattachments', [])

                # normalize
                for assign_file in assign_files:
                    file_type = assign_file.get('type', '')
                    if file_type is None or file_type == '':
                        assign_file.update({'type': 'assign_file'})

                course_assigns.update(
                    {
                        assign_id: {
                            'id': assign_rid,
                            'files': assign_files,
                            'name': assign_name,
                            'timemodified': assign_timemodified,
                        }
                    }
                )

            result.update({course_id: course_assigns})

        return result

    def fetch_submissions(self, userid: int, assignments: {int: {int: {}}}) -> {int: {int: {}}}:
        """
        Fetches for the assignments list of all courses the additionally
        submissions. This is kind of waste of resources, because there
        is no API to get all submissions at once
        @param userid: the user id.
        @param assignments: the dictionary of assignments of all courses.
        @return: A Dictionary of all assignments,
                 indexed by courses, then assignment
        """
        # do this only if version is greater then 3.1
        # because mod_assign_get_submission_status will fail
        if self.version < 2016052300:
            return assignments

        intro = '\rDownloading submission information'

        counter = 0
        total = 0

        # count total assignments for nice console output
        for course_id in assignments:
            for assignment_id in assignments[course_id]:
                total += 1

        for course_id in assignments:
            for assignment_id in assignments[course_id]:
                counter += 1
                assign = assignments[course_id][assignment_id]
                real_id = assign.get('id', 0)
                data = {'userid': userid, 'assignid': real_id}

                shorted_assign_name = assign.get('name', '')
                if len(shorted_assign_name) > 17:
                    shorted_assign_name = shorted_assign_name[:15] + '..'

                print(intro + ' %3d/%3d [%-17s|%6s]\033[K' % (counter, total, shorted_assign_name, course_id), end='')

                submission = self.request_helper.post_REST('mod_assign_get_submission_status', data)

                submission_files = self._get_files_of_submission(submission)
                assign['files'] += submission_files

        return assignments

    @staticmethod
    def _get_files_of_submission(submission: {}) -> []:
        result = []
        # get own submissions
        lastattempt = submission.get('lastattempt', {})
        l_submission = lastattempt.get('submission', {})
        l_teamsubmission = lastattempt.get('teamsubmission', {})

        # get teachers feedback
        feedback = submission.get('feedback', {})

        result += AssignmentsHandler._get_files_of_plugins(l_submission)
        result += AssignmentsHandler._get_files_of_plugins(l_teamsubmission)
        result += AssignmentsHandler._get_files_of_plugins(feedback)
        result += AssignmentsHandler._get_grade_of_feedback(feedback)

        return result

    @staticmethod
    def _get_grade_of_feedback(feedback: {}) -> []:
        result = []

        gradefordisplay = feedback.get('gradefordisplay', "")
        gradeddate = feedback.get('gradeddate', 0)
        if gradeddate is None or gradefordisplay is None or gradeddate == 0 or gradefordisplay == "":
            return result

        file = {
            'filename': 'grade',
            'filepath': '/',
            'timemodified': gradeddate,
            'description': gradefordisplay,
            'type': 'description',
        }

        result.append(file)

        return result

    @staticmethod
    def _get_files_of_plugins(obj: {}) -> []:
        result = []
        plugins = obj.get('plugins', [])

        for plugin in plugins:
            fileareas = plugin.get('fileareas', [])

            for filearea in fileareas:
                files = filearea.get('files', [])

                for file in files:
                    file_type = file.get('type', '')
                    if file_type is None or file_type == '':
                        file.update({'type': 'submission_file'})

                    result.append(file)

        for plugin in plugins:
            editorfields = plugin.get('editorfields', [])

            for editorfield in editorfields:

                filename = editorfield.get('description', '')
                description = editorfield.get('text', '')
                if filename != '' and description != '':
                    description_file = {
                        'filename': filename,
                        'description': description,
                        'type': 'description',
                    }
                    result.append(description_file)

        return result



# coding: utf-8
from __future__ import unicode_literals


import re
import json

from youtube_dl.compat import (
    compat_urllib_parse,
    compat_urllib_parse_urlparse,
)

from youtube_dl.extractor.common import InfoExtractor
from youtube_dl.utils import (
    ExtractorError,
    urlencode_postdata,
    js_to_json,
    determine_ext,
    mimetype2ext,
    extract_attributes,
    HEADRequest,
)


class Helixmedia(InfoExtractor):
    IE_NAME = 'helixmedia'
    _VALID_URL = r'(?P<scheme>https?://)(?P<host>[^/]+)(?P<path>.*)?/mod/helixmedia/view.php\?.*?id=(?P<id>\d+)'
    _LAUNCH_FORM = 'ltiLaunchForm'

    # _TEST = {'url': 'http://localhost/moodle/mod/helixmedia/view.php?id=3'}

    def _real_extract(self, url):
        mobj = re.match(self._VALID_URL, url)
        scheme = mobj.group('scheme')
        host = mobj.group('host')
        path = mobj.group('path')
        video_id = mobj.group('id')

        launch_url = scheme + host + path + '/mod/helixmedia/launch.php?type=1&id=' + video_id

        # webpage = self._download_webpage(url, video_id)
        launch_webpage = self._download_webpage(launch_url, video_id, 'Downloading helixmedia launch webpage')
        launch_inputs = self._form_hidden_inputs(self._LAUNCH_FORM, launch_webpage)
        launch_form_str = self._search_regex(
            r'(?P<form><form[^>]+?id=(["\'])%s\2[^>]*>)' % self._LAUNCH_FORM, launch_webpage, 'login form', group='form'
        )

        action_url = extract_attributes(launch_form_str).get('action')

        submit_page, start_urlh = self._download_webpage_handle(
            action_url, video_id, 'Launch helixmedia app', data=urlencode_postdata(launch_inputs)
        )

        if 'UploadSessionId' not in start_urlh.geturl():
            raise ExtractorError('Unable to launch helixmedia video', expected=True)

        parsed_mediaserver_url = list(compat_urllib_parse_urlparse(start_urlh.geturl()))
        parsed_mediaserver_url[4] += '&mobile=N&fullWidth=940&fullHeight=906'
        parsed_mediaserver_url[2] += 'Split'
        mediaserver_url = compat_urllib_parse.urlunparse(parsed_mediaserver_url)

        video_webpage = self._download_webpage(mediaserver_url, video_id, 'Downloading video details')

        video_model = json.loads(js_to_json(self._search_regex(r'var model = ([^;]+);', video_webpage, 'video model')))

        video_title = video_model.get('VideoTitle', None)

        video_description = video_model.get('VideoDescription', '')
        video_id = str(video_model.get('VideoId', video_id))
        download_url = video_model.get('DownloadUrl', None)
        video_json = json.loads(video_model.get('PlayScreenVm', {}).get('VodPlayerModel', {}).get('PlayerJson', '{}'))
        if video_json == {}:
            video_json = json.loads(video_model.get('VodPlayerModel', {}).get('PlayerJson', '{}'))
        if video_title is None:
            video_title = video_json.get('abouttext', 'Unknown title')

        thumbnail_list = video_json.get('tracks', [])
        thumbnail = None
        if len(thumbnail_list) >= 1:
            thumbnail = thumbnail_list[0].get('file', None)
            if thumbnail is not None:
                thumbnail = thumbnail.replace('vtt', 'jpg')

        sources_list = video_json.get('sources', [])

        formats = []
        for track in sources_list:
            href = track['file']
            ext = determine_ext(href, None)

            if ext == 'mpd':
                # DASH
                formats.extend(self._extract_mpd_formats(href, video_id, mpd_id='dash', fatal=False))
            elif ext == 'm3u8':
                # HLS
                formats.extend(
                    self._extract_m3u8_formats(href, video_id, m3u8_id='hls', entry_protocol='m3u8_native', fatal=False)
                )
            elif ext == 'f4m':
                # HDS
                formats.extend(self._extract_f4m_formats(href, video_id, f4m_id='hds', fatal=False))
            elif ext == 'smil':
                formats.extend(self._extract_smil_formats(href, video_id, fatal=False))
            else:
                track_obj = {
                    'url': href,
                    'ext': ext,
                }
                formats.append(track_obj)

        if download_url is not None:
            ext_req = HEADRequest(download_url)
            ext_handle = self._request_webpage(ext_req, video_id, note='Determining extension')
            ext = self.urlhandle_detect_ext(ext_handle)

            track_obj_direct = {
                'url': download_url,
                'ext': ext,
            }
            formats.append(track_obj_direct)

        self._sort_formats(formats)

        result_obj = {'formats': formats}

        if video_id is not None:
            result_obj.update({'id': video_id})

        if video_title is not None:
            result_obj.update({'title': video_title})

        if video_description is not None:
            result_obj.update({'creator': video_description})

        if thumbnail is not None:
            result_obj.update({'thumbnail': thumbnail})

        return result_obj

    def urlhandle_detect_ext(self, url_handle):
        getheader = url_handle.headers.get

        cd = getheader('Content-Disposition')
        if cd:
            m = re.match(r'attachment;.*filename="(?P<filename>[^"]+)"', cd)
            if m:
                e = determine_ext(m.group('filename'), default_ext=None)
                if e:
                    return e

        return mimetype2ext(getheader('Content-Type'))



import re
import sys
import base64

from http.server import BaseHTTPRequestHandler, HTTPServer

from moodle_dl.utils.logger import Log


class TransferServer(BaseHTTPRequestHandler):
    """
    Transfer server is an HTTP-server that waits for an incoming SSO token.
    """

    received_token = ''

    def do_GET(self):
        if self.path == '/favicon.ico':
            # Don't serve favicon
            self.send_response(404)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            return

        TransferServer.received_token = self.path

        # Token recieved
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(('Moodle Downloader - Token was successfully transferred.').encode('utf-8'))

    def log_message(self, format, *args):
        return


def extract_token(address: str) -> str:
    """
    Extracts a token from a returned URL
    """
    splitted = address.split('token=')

    if len(splitted) < 2:
        return None

    decoded = str(base64.b64decode(splitted[1]))

    splitted = decoded.split(':::')
    if len(splitted) < 2:
        return None

    token = re.sub(r'[^A-Za-z0-9]+', '', splitted[1])

    if len(splitted) < 3:
        return token, None
    else:
        secret_token = re.sub(r'[^A-Za-z0-9]+', '', splitted[2])
        return token, secret_token


def receive_token() -> str:
    """
    Starts an HTTP server to receive the SSO token from browser.
    It waits till a token was received.
    """
    server_address = ('localhost', 80)
    try:
        httpd = HTTPServer(server_address, TransferServer)
    except PermissionError:
        Log.error(
            'Permission denied: Please start the'
            + ' downloader once with administrator rights, so that it'
            + ' can wait on port 80 for the token.'
        )
        sys.exit(1)

    extracted_token = None

    while extracted_token is None:
        httpd.handle_request()

        extracted_token = extract_token(TransferServer.received_token)

    httpd.server_close()

    return extracted_token



from datetime import datetime

from moodle_dl.moodle_connector.request_helper import RequestHelper
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.path_tools import PathTools


class ForumsHandler:
    """
    Fetches and parses the various endpoints in Moodle for Forum Entries.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_forums(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Databases List for all courses from the
        Moodle system
        @return: A Dictionary of all databases,
                 indexed by courses, then databases
        """
        # do this only if version is greater then 2.5
        # because mod_forum_get_forums_by_courses will fail
        if self.version < 2013051400:
            return {}

        print('\rDownloading forums information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        forums = self.request_helper.post_REST('mod_forum_get_forums_by_courses', extra_data)

        result = {}
        for forum in forums:
            # This is the instance id with which we can make the API queries.
            forum_id = forum.get('id', 0)
            forum_name = forum.get('name', 'forum')
            forum_intro = forum.get('intro', '')
            forum_course_module_id = forum.get('cmid', 0)
            forum_introfiles = forum.get('introfiles', [])
            course_id = forum.get('course', 0)

            # normalize
            for forum_file in forum_introfiles:
                file_type = forum_file.get('type', '')
                if file_type is None or file_type == '':
                    forum_file.update({'type': 'forum_introfile'})

            forum_entry = {
                forum_course_module_id: {
                    'id': forum_id,
                    'name': forum_name,
                    'intro': forum_intro,
                    'files': forum_introfiles,
                }
            }

            course_dic = result.get(course_id, {})

            course_dic.update(forum_entry)

            result.update({course_id: course_dic})

        return result

    def fetch_forums_posts(self, forums: {}, last_timestamps_per_forum: {}) -> {}:
        """
        Fetches for the forums list of all courses the additionally
        entries. This is kind of waste of resources, because there
        is no API to get all entries at once.
        @param forums: the dictionary of forums of all courses.
        @return: A Dictionary of all forums,
                 indexed by courses, then forums
        """
        # do this only if version is greater then 2.8
        # because mod_forum_get_forum_discussions_paginated will fail
        if self.version < 2014111000:
            return forums

        counter = 0
        total = 0
        # count total forums for nice console output
        for course_id in forums:
            for forum_id in forums[course_id]:
                total += 1

        for course_id in forums:
            for forum_id in forums[course_id]:
                counter += 1
                real_id = forums[course_id][forum_id].get('id', 0)
                page_num = 0
                last_timestamp = last_timestamps_per_forum.get(forum_id, 0)
                latest_discussions = []
                done = False
                while not done:
                    data = {
                        'forumid': real_id,
                        'perpage': 10,
                        'page': page_num,
                    }

                    print(
                        '\rDownloading forum discussions %3d/%3d [%6s|%6s|p%s]\033[K'
                        % (counter, total, course_id, real_id, page_num),
                        end='',
                    )

                    if self.version >= 2019052000:
                        discussions_result = self.request_helper.post_REST('mod_forum_get_forum_discussions', data)
                    else:
                        discussions_result = self.request_helper.post_REST(
                            'mod_forum_get_forum_discussions_paginated', data
                        )

                    discussions = discussions_result.get('discussions', [])

                    if len(discussions) == 0:
                        done = True
                        break

                    for discussion in discussions:

                        timemodified = discussion.get('timemodified', 0)
                        if discussion.get('modified', 0) > timemodified:
                            timemodified = discussion.get('modified', 0)

                        if last_timestamp < timemodified:
                            latest_discussions.append(
                                {
                                    'subject': discussion.get('subject', ''),
                                    'timemodified': timemodified,
                                    'discussion_id': discussion.get('discussion', 0),
                                    'created': discussion.get('created', 0),
                                }
                            )
                        else:
                            done = True
                            break
                    page_num += 1

                forums_files = self._get_files_of_discussions(latest_discussions)
                forums[course_id][forum_id]['files'] += forums_files

        return forums

    def _get_files_of_discussions(self, latest_discussions: []) -> []:
        result = []

        for i, discussion in enumerate(latest_discussions):
            valid_subject = PathTools.to_valid_name(discussion.get('subject', ''))
            shorted_discussion_name = valid_subject
            if len(shorted_discussion_name) > 17:
                shorted_discussion_name = shorted_discussion_name[:15] + '..'
            discussion_id = discussion.get('discussion_id', 0)
            discussion_created = discussion.get('created', 0)

            print(
                '\rDownloading posts of discussion [%-17s|%6s] %3d/%3d\033[K'
                % (shorted_discussion_name, discussion_id, i, len(latest_discussions) - 1),
                end='',
            )

            data = {
                'discussionid': discussion_id,
                'sortby': 'modified',
                'sortdirection': 'ASC',
            }

            posts_result = self.request_helper.post_REST('mod_forum_get_forum_discussion_posts', data)

            posts = posts_result.get('posts', [])

            for post in posts:
                post_message = post.get('message', '')
                if post_message is None:
                    post_message = ''
                post_modified = post.get('modified', 0)

                post_id = post.get('id', 0)
                post_parent = post.get('parent', 0)
                post_userfullname = post.get('userfullname', '')
                if post_userfullname is None:
                    post_userfullname = "Unknown"

                post_filename = PathTools.to_valid_name('[' + str(post_id) + '] ' + post_userfullname)
                if post_parent != 0:
                    post_filename = PathTools.to_valid_name(post_filename + ' response to [' + str(post_parent) + ']')

                post_path = PathTools.to_valid_name(
                    datetime.utcfromtimestamp(discussion_created).strftime('%y-%m-%d') + ' ' + valid_subject
                )

                post_files = post.get('messageinlinefiles', [])
                post_files += post.get('attachments', [])

                post_file = {
                    'filename': post_filename,
                    'filepath': post_path,
                    'timemodified': post_modified,
                    'description': post_message,
                    'type': 'description',
                }
                result.append(post_file)

                for post_file in post_files:
                    file_type = post_file.get('type', '')
                    if file_type is None or file_type == '':
                        post_file.update({'type': 'forum_file'})
                    post_file.update({'filepath': post_path})
                    result.append(post_file)

        return result
